package dev.pfaff.altargui.libimage;

import dev.pfaff.altargui.GuiGlobals;
import dev.pfaff.log4truth.Logger;
import dev.pfaff.log4truth.format.Format;

import java.lang.foreign.*;
import java.lang.invoke.MethodHandle;
import java.lang.invoke.VarHandle;
import java.util.List;

import static dev.pfaff.altargui.util.MethodHandleUtil.withoutBase;
import static dev.pfaff.log4truth.StandardTags.DEBUG;
import static java.lang.foreign.MemoryLayout.PathElement.groupElement;
import static java.lang.foreign.MemoryLayout.structLayout;
import static java.lang.foreign.ValueLayout.*;
import static java.lang.invoke.MethodHandles.explicitCastArguments;
import static java.lang.invoke.MethodType.methodType;

public final class LibImage {
	private static final Linker LINKER = Linker.nativeLinker();
	private static final SymbolLookup LOOKUP = GuiGlobals.HOST.loadLibImage();

	private static final ValueLayout ML_ColorType = JAVA_BYTE.withName(ColorType.class.getSimpleName());
	private static final StructLayout ML_ImageInfo = structLayout(JAVA_INT.withName("width"),
			JAVA_INT.withName("height"),
			JAVA_LONG.withName("byte_count"),
			ML_ColorType.withName("color_type"));
	private static final VarHandle VH_ImageInfo_width =
			withoutBase(ML_ImageInfo.varHandle(groupElement("width")).withInvokeExactBehavior());
	private static final VarHandle VH_ImageInfo_height =
			withoutBase(ML_ImageInfo.varHandle(groupElement("height")).withInvokeExactBehavior());
	private static final VarHandle VH_ImageInfo_byte_count =
			withoutBase(ML_ImageInfo.varHandle(groupElement("byte_count")).withInvokeExactBehavior());
	private static final VarHandle VH_ImageInfo_color_type =
			withoutBase(ML_ImageInfo.varHandle(groupElement("color_type")).withInvokeExactBehavior());

	private static final ValueLayout ML_PtrInt = ADDRESS.byteSize() == JAVA_LONG.byteSize() ? JAVA_LONG : JAVA_INT;

	// TODO: these should not be critical, but we need to for now to avoid copying the bytes into a native memory
	//  segment.
	private static final MethodHandle MH_load_png_info = explicitCastArguments(LINKER.downcallHandle(LOOKUP.find(
							"load_png_info").get(),
					FunctionDescriptor.of(JAVA_BOOLEAN.withName("errored"),
							ADDRESS.withTargetLayout(JAVA_BYTE).withName("inputPtr"),
							ML_PtrInt.withName("inputLen"),
							ADDRESS.withTargetLayout(ML_ImageInfo).withName("info")),
					Linker.Option.critical(true)),
			methodType(boolean.class, MemorySegment.class, long.class, MemorySegment.class));
	private static final MethodHandle MH_load_png = explicitCastArguments(LINKER.downcallHandle(LOOKUP.find(
							"load_png").get(),
					FunctionDescriptor.of(JAVA_BOOLEAN.withName("errored"),
							ADDRESS.withTargetLayout(JAVA_BYTE).withName("inputPtr"),
							ML_PtrInt.withName("inputLen"),
							ADDRESS.withTargetLayout(JAVA_BYTE).withName("outputPtr"),
							ML_PtrInt.withName("outputLen")),
					Linker.Option.critical(true)),
			methodType(boolean.class, MemorySegment.class, long.class, MemorySegment.class, long.class));

	public static ImageInfo loadPngInfo(MemorySegment bytes) {
		long start = System.nanoTime();
		try (var arena = Arena.ofConfined()) {
			var result = arena.allocate(ML_ImageInfo);
			if ((boolean) MH_load_png_info.invokeExact(bytes, bytes.byteSize(), result)) {
				throw new RuntimeException("Failed to load png info");
			}
			return new ImageInfo(
					(int) VH_ImageInfo_width.get(result),
					(int) VH_ImageInfo_height.get(result),
					(long) VH_ImageInfo_byte_count.get(result),
					ColorType.VALUES.get((byte) VH_ImageInfo_color_type.get(result))
			);
		} catch (Throwable e) {
			throw new RuntimeException(e);
		} finally {
			long elapsed = System.nanoTime() - start;
			Logger.log(() -> "Completed loadPngInfo in " + Format.STANDARD_DURATION.nanosToString(elapsed),
					List.of(DEBUG));
		}
	}

	public static void loadPng(MemorySegment bytes, MemorySegment pixels) {
		long start = System.nanoTime();
		try {
			if ((boolean) MH_load_png.invokeExact(bytes, bytes.byteSize(), pixels, pixels.byteSize())) {
				throw new RuntimeException("Failed to load png info");
			}
		} catch (Throwable e) {
			throw new RuntimeException(e);
		} finally {
			long elapsed = System.nanoTime() - start;
			Logger.log(() -> "Completed loadPng in " + Format.STANDARD_DURATION.nanosToString(elapsed),
					List.of(DEBUG));
		}
	}

	public static ImageData loadPng(MemorySegment bytes, Arena pixelsArena) {
		var info = loadPngInfo(bytes);
		var pixels = pixelsArena.allocate(info.byteCount(), 8);
		loadPng(bytes, pixels);
		return new ImageData(info, pixels);
	}
}
