package dev.pfaff.altargui.widgets;

import dev.pfaff.altargui.StateContext;
import dev.pfaff.altargui.Widget;
import dev.pfaff.altargui.WidgetInstance;
import dev.pfaff.altargui.WidgetInstantiationArgs;
import dev.pfaff.altargui.element.StatefulWidgetInstance;
import dev.pfaff.altargui.handlers.CharTypedHandler;
import dev.pfaff.altargui.handlers.KeyPressHandler;
import dev.pfaff.altargui.handlers.MouseClickHandler;
import dev.pfaff.altargui.input.InputKey;
import dev.pfaff.altargui.input.InputMod;
import dev.pfaff.altargui.input.InputMods;
import dev.pfaff.altargui.types.Color;
import org.jspecify.annotations.NonNull;
import org.jspecify.annotations.Nullable;

import java.lang.ref.WeakReference;
import java.util.List;
import java.util.Objects;
import java.util.function.Consumer;

import static dev.pfaff.altargui.GuiGlobals.HOST;
import static dev.pfaff.altargui.widgets.MultiChildWidget.row;

public record TextFieldWidget(@Nullable TextFieldController controller, @NonNull Consumer<String> onChanged) implements Widget {

	public static final int LINE_HEIGHT = 9;

	public TextFieldWidget() {
		this(null, ignored -> {});
	}

	public TextFieldWidget withController(@Nullable TextFieldController controller) {
		return new TextFieldWidget(controller, onChanged);
	}

	public TextFieldWidget withOnChanged(@NonNull Consumer<String> onChanged) {
		return new TextFieldWidget(controller, onChanged);
	}

	@Override
	public @NonNull WidgetInstance<TextFieldWidget> instantiate(WidgetInstantiationArgs args) {
		return new State(args);
	}

	private static final class State extends StatefulWidgetInstance<TextFieldWidget> implements CharTypedHandler, KeyPressHandler, MouseClickHandler {
		private boolean focused;
		private TextFieldController controller;

		private State(WidgetInstantiationArgs args) {
			super(args);
		}

		@Override
		public void configure(StateContext stateContext, TextFieldWidget widget, boolean init) {
			boolean dirty = init;
			if (init || (widget.controller != null && widget.controller != controller)) {
				controller = Objects.requireNonNullElseGet(widget.controller, TextFieldController::new);
			}
			if (!controller.currentState.refersTo(this)) {
				dirty = true;
				controller.currentState = new WeakReference<>(this);
			}
			controller.currentListener = widget.onChanged;
			if (dirty) markNeedsRebuild();
		}

		@Override
		protected @NonNull Widget build() {
			var v = controller.buffer.toString();
			Widget widget;
			if (focused) {
				int cursorI = controller.cursor;
				int selectionExtent = controller.selectionExtent;
				int selectionStart = selectionExtent == -1 ? cursorI : Math.min(cursorI, selectionExtent);
				int selectionEnd = selectionExtent == -1 ? cursorI : Math.max(cursorI, selectionExtent);
				var cursor = RectWidget.colored(Color.WHITE)
						.withAbsoluteWidth(1)
						.atRelativePosition(-0.5f, 0f)
						.withAbsoluteWidth(cursorI == 0 || cursorI == v.length() ? 0 : 1);
				Widget widget1, widget2;
				if (selectionExtent != -1) {
					var selected = TextWidget.of(v.substring(selectionStart, selectionEnd))
							.withBackgroundColor(Color.fromRgba(0x0090ff_0f));
					widget1 = cursorI < selectionExtent ? cursor : selected;
					widget2 = cursorI < selectionExtent ? selected : cursor;
				} else {
					widget1 = cursor;
					widget2 = Widgets.EMPTY;
				}
				widget = row(List.of(TextWidget.of(v.substring(0, selectionStart)),
								   widget1,
								   widget2,
								   TextWidget.of(v.substring(selectionEnd))));
			} else {
				widget = row(List.of(TextWidget.of(v)));
			}
			widget = widget.withAbsoluteHeight(LINE_HEIGHT).withMargin(2);
			widget = widget.withBackground(focused ? RectWidget.GENERIC_HEAVY_HUD_BACKGROUND : RectWidget.GENERIC_HUD_BACKGROUND)
					.withBorder(Color.fromArgb(focused ? 0x99_bcbcbc : 0x99_222222), 2f);
			return widget;
		}

		@Override
		public boolean charTyped(int codePoint) {
			if (!focused) return false;
			return controller.insertAtCursor(codePoint);
		}

		@Override
		public boolean keyPressed(InputKey keyCode, int scanCode, InputMods modifiers) {
			if (!focused) return false;
			boolean hasPrimary = modifiers.has(InputMod.PRIMARY);
			boolean hasShift = modifiers.has(InputMod.SHIFT);
			switch (keyCode) {
				case TAB -> {
					// TODO: autocompletion
				}
				case ENTER -> {
					// TODO: onSubmit
				}
				case LEFT -> {
					controller.extendSelection(-1, hasShift);
					return true;
				}
				case RIGHT -> {
					controller.extendSelection(+1, hasShift);
					return true;
				}
				case HOME -> {
					controller.selectTo(0, hasShift);
					return true;
				}
				case END -> {
					controller.selectTo(controller.buffer.length(), hasShift);
					return true;
				}
				case BACKSPACE -> {
					controller.backspace(hasPrimary);
					return true;
				}
				case DELETE -> {
					controller.delete(hasPrimary);
					return true;
				}
				case V -> {
					if (hasPrimary) {
						controller.pasteAsync();
						return true;
					}
				}
				case X, C -> {
					if (hasPrimary) {
						controller.copyAsync(keyCode == InputKey.X);
						return true;
					}
				}
			}

			return false;
		}

		@Override
		public boolean mouseClicked(double mouseX, double mouseY, int button) {
			if (button == 0 && isInside(mouseX, mouseY)) {
				if (!focused) {
					focused = true;
					markNeedsRebuild();
				}
				return true;
			} else {
				if (focused) {
					focused = false;
					markNeedsRebuild();
				}
				return false;
			}
		}

		private void markNeedsRebuildAccessible() {
			this.markNeedsRebuild();
		}
	}

	// TODO: consider implementing Listenable
	public static final class TextFieldController {
		private static final WeakReference<@Nullable State> NULL_STATE = new WeakReference<>(null);

		private WeakReference<State> currentState = NULL_STATE;
		private Consumer<String> currentListener;

		private final @NonNull StringBuilder buffer;

		private int cursor;
		private int selectionExtent;

		public TextFieldController() {
			buffer = new StringBuilder();
		}

		public TextFieldController(@NonNull String initialValue) {
			buffer = new StringBuilder(initialValue);
		}

		public String value() {
			return buffer.toString();
		}

		public void value(String s) {
			buffer.setLength(0);
			buffer.append(s);
			markDirty(false);
		}

		public boolean selectTo(int position, boolean isShiftDown) {
			if (position == cursor) return false;

			if (position >= 0 && position <= buffer.length()) {
				if (isShiftDown && selectionExtent == -1) {
					selectionExtent = cursor;
				}

				cursor = position;
				markDirty(true);
				return true;
			}
			return false;
		}

		public boolean extendSelection(int delta, boolean isShiftDown) {
			if (!isShiftDown && selectionExtent != -1) {
				if (delta == -1) {
					cursor = Math.min(cursor, selectionExtent);
					selectionExtent = -1;
					markDirty(true);
					return true;
				} else if (delta == 1) {
					cursor = Math.max(cursor, selectionExtent);
					selectionExtent = -1;
					markDirty(true);
					return true;
				}
			}
			return selectTo(cursor + delta, isShiftDown);
		}

		private boolean deleteSelection() {
			if (selectionExtent != -1 && selectionExtent != cursor) {
				buffer.delete(selectionStart(), selectionEnd());
				cursor = selectionStart();
				selectionExtent = -1;
				markDirty(true);
				return true;
			}
			return false;
		}

		public boolean backspace(boolean isCtrlDown) {
			if (cursor == 0) return false;
            if (isCtrlDown) {
                buffer.delete(0, cursor);
                selectionExtent = -1;
                cursor = 0;
				markDirty(true);
                return true;
            } else {
                if (selectionExtent != -1) {
                    return deleteSelection();
                } else if (cursor > 0) {
                    buffer.deleteCharAt(--cursor);
					markDirty(true);
                    return true;
                }
            }
			return false;
        }

		public boolean delete(boolean isCtrlDown) {
            if (buffer.isEmpty()) return false;
            if (isCtrlDown) {
                var len = buffer.length();
                buffer.delete(cursor, len);
                selectionExtent = -1;
				markDirty(true);
                return true;
            } else {
                if (selectionExtent != -1) {
                    deleteSelection();
					markDirty(true);
                    return true;
                } else if (cursor < buffer.length()) {
                    buffer.deleteCharAt(cursor);
					markDirty(true);
                    return true;
                }
            }
            return false;
		}

		public int selectionStart() {
			return Math.min(cursor, selectionExtent);
		}

		public int selectionEnd() {
			return Math.max(cursor, selectionExtent);
		}

		public String getSelection() {
			if (selectionExtent == -1 || selectionExtent == cursor) return "";
			return buffer.substring(selectionStart(), selectionEnd());
		}

		public void pasteAsync() {
			Thread.startVirtualThread(() -> {
				var s = HOST.getClipboardString();
				if (s == null) return;
				if (selectionStart() != -1) {
					buffer.replace(selectionStart(),
											  selectionEnd(),
											  s);
					cursor = selectionStart() + s.length();
					selectionExtent = -1;
					markDirty(true);
				} else if (!s.isEmpty()) {
					buffer.insert(cursor, s);
					cursor += s.length();
					markDirty(true);
				}
			});
		}

		public void copyAsync(boolean cut) {
			var s = getSelection();
			if (!s.isEmpty()) {
				Thread.startVirtualThread(() -> HOST.setClipboardString(s));
				if (cut) {
					buffer.delete(selectionStart(), selectionEnd());
					cursor = selectionStart();
					selectionExtent = -1;
					markDirty(true);
				}
			}
		}

		private void markDirty(boolean notify) {
			var state = currentState.get();
			if (state != null) state.markNeedsRebuildAccessible();
			if (notify) currentListener.accept(value());
		}

		public boolean insertAtCursor(int codePoint) {
			if (!Character.isDefined(codePoint)) return false;
			if (Character.isISOControl(codePoint)) return false;

			// TODO: make sure we avoid all control characters.
			if (selectionExtent != -1) {
				buffer.replace(selectionStart(),
						selectionEnd(),
						Character.toString(codePoint));
				cursor = selectionStart() + 1;
				selectionExtent = -1;
			} else {
				buffer.insert(cursor++, Character.toString(codePoint));
			}
			markDirty(true);
			return true;
		}

		public boolean insertAtCursor(char chr) {
			return insertAtCursor((int) chr);
		}
	}
}
